; -------------------------------------------------------------
; This file contains a (incomplete) collection of MASM 6+
; macros, equates, and data and procedure definitions aimed
; at supporting real-mode DOS programs.
;
; This file should be included at the start of your source.
;
; The procedures depend on DS being set to the data segment.
;
; These procedures preserve any segment register they alter.
;
; The procedures (are supposed to) preserve all registers
; other than those used to return a value.
;
; Defining the symbol DEBUG_ON will cause the print macro to
; preserve the 16-bit flags and the 32-bit registers across
; the call. The dumpflags and dumpmem procedures are intended
; primarily for debugging, so they always perserve the 16-bit
; flags and the 32-bit registers across the call.
;
; The advantage of using STDIN and STDOUT is that input and
; output can be redirected.
;
; Most recent update 11-6-07:
;
; Added code to preserve AX around invoke in the putch macro.
;
; Removed duplicate of dword$ macro.
;
; Corrected various comments.
; -------------------------------------------------------------

    ; -----------------------------------------------
    ; DOS handles for standard input/output devices.
    ; -----------------------------------------------

    HSTDIN   EQU 0
    HSTDOUT  EQU 1

    ; ---------------------------------------------------
    ; The prototypes are not actually necessary, but are
    ; included primarily to serve as a syntax reference.
    ; ---------------------------------------------------

    stdin     PROTO c npdest:WORD, blen:WORD
    szlen     PROTO c npsz:WORD
    stdout    PROTO c npsz:WORD
    putchar   PROTO c char:BYTE
    waitkey   PROTO
    locate    PROTO c row:BYTE, col:BYTE
    atodw     PROTO c npsz:WORD
    b2dec     PROTO c src:BYTE, npdest:WORD
    b2hex     PROTO c src:BYTE, npdest:WORD
    w2dec     PROTO c src:WORD, npdest:WORD
    w2hex     PROTO c src:WORD, npdest:WORD
    dw2dec    PROTO c src:DWORD, npdest:WORD
    dw2hex    PROTO c src:DWORD, npdest:WORD
    b2bin     PROTO c src:BYTE, npdest:WORD
    szlower   PROTO c npsz:WORD
    szupper   PROTO c npsz:WORD

    dumpflags PROTO
    dumpmem   PROTO mem_seg:WORD, mem_off:WORD, paragraphs:WORD

    get_tick_count   PROTO c
    delay            PROTO c dwcount:DWORD
    read_cmos        PROTO c address:BYTE
    print_far_string PROTO c str_seg:WORD, str_ptr:WORD

    ; -----------------------------------------------------
    ; This is the MASM32 reparg macro with *no* changes.
    ;
    ; "This macro replaces quoted text with a DATA section
    ; OFFSET and returns it in ADDR "name" format."
    ; -----------------------------------------------------

    reparg MACRO arg
      LOCAL nustr
      quot SUBSTR <arg>, 1, 1
      IFIDN quot,<">          ;; if 1st char = "
        .data
          nustr db arg, 0     ;; write arg to .DATA section
        .code
        EXITM <ADDR nustr>    ;; append name to ADDR operator
      ELSE
        EXITM <arg>           ;; else return arg
      ENDIF
    ENDM

    ; -----------------------------------------------------
    ; This is a VARARG version of the MASM32 print macro.
    ; It will accept either a quoted string, or the offset
    ; address of a null-terminated string in the data
    ; segment.
    ; -----------------------------------------------------

    print MACRO varname:VARARG
      IFDEF DEBUG_ON
        pushf
        pushad
      ENDIF
      FOR arg, <varname>
        invoke stdout, reparg(arg)
      ENDM
      IFDEF DEBUG_ON
        popad
        popf
      ENDIF
    ENDM

    ; ------------------------------------------------------
    ; This is the MASM32 chr$ macro with *no* changes.
    ;
    ; "This macro takes quoted string and ascii numbers in
    ; the same format as a .DATA section entry. It returns
    ; the offset in the .DATA section of the data contained
    ; in its parameters."
    ; ------------------------------------------------------

    chr$ MACRO any_text:VARARG
      LOCAL txtname
      .data
        txtname db any_text,0
      .code
      EXITM <OFFSET txtname>
    ENDM

    ; ----------------------------------------------------
    ; These macros make it easier to print signed decimal
    ; and unsigned hex values.
    ;
    ; Note that the decimal conversion procedures cannot
    ; correctly handle out of range values.
    ;
    ; Note also that the conversions reuse a single output
    ; buffer, so if more then one conversion is placed in
    ; a single print statement all of the conversions will
    ; effectively return the value of the last conversion.
    ;
    ; The pushs and pops in the byte versions are because
    ; for a byte register argument other than AL invoke
    ; uses AL for a scratch register.
    ; ----------------------------------------------------

    byte$ MACRO src
      push ax
      invoke b2dec,src,ADDR outBuf
      pop ax
      EXITM <ADDR outBuf>
    ENDM

    hexbyte$ MACRO src
      push ax
      invoke b2hex,src,ADDR outBuf
      pop ax
      EXITM <ADDR outBuf>
    ENDM

    word$ MACRO src
      invoke w2dec,src,ADDR outBuf
      EXITM <ADDR outBuf>
    ENDM

    hexword$ MACRO src
      invoke w2hex,src,ADDR outBuf
      EXITM <ADDR outBuf>
    ENDM

    dword$ MACRO src
      invoke dw2dec,src,ADDR outBuf
      EXITM <ADDR outBuf>
    ENDM

    hexdword$ MACRO src
      invoke dw2hex,src,ADDR outBuf
      EXITM <ADDR outBuf>
    ENDM

    ; -------------------------------------------------
    ; This macro makes it easier to print characters.
    ; AX must be preserved around invoke because for
    ; a byte register argument other than AL invoke
    ; uses AL for a scratch register.
    ; -------------------------------------------------

    putch MACRO char
      push ax
      invoke putchar, char
      pop ax
    ENDM

    cls MACRO
      call _cls
    ENDM

    ; ----------------------------------------------------
    ; This macro sets the cursor position for display
    ; page 0 to the specified row and column coordinates.
    ; The unusual feature is that either or both of the
    ; coordinate parameters may be left blank, causing
    ; the cursor position with respect to the blank
    ; coordinate to remain unchanged.
    ; ----------------------------------------------------

    loc MACRO row,col
      IFNB <row>
        IFNB <col>
          invoke locate, row, col
        ELSE
          call cursorcol
          invoke locate, row, al
        ENDIF
      ELSE
        IFNB <col>
          call cursorrow
          invoke locate, al, col
        ELSE
          call cursorrow
          mov   ah, al
          call cursorcol
          invoke locate, ah, al
        ENDIF
      ENDIF
    ENDM

    ; -------------------------------------------------
    ; These macros return a bit range value, with the
    ; value shifted to place lowbit at bit position 0.
    ; -------------------------------------------------

    brval8 MACRO byteval, highbit, lowbit
      mov al, byteval
      shl al, 7-highbit
      shr al, 7-highbit+lowbit
      EXITM <al>
    ENDM

    brval16 MACRO wordval, highbit, lowbit
      mov ax, wordval
      shl ax, 15-highbit
      shr ax, 15-highbit+lowbit
      EXITM <ax>
    ENDM

    brval32 MACRO dwordval, highbit, lowbit
      mov eax, dwordval
      shl eax, 31-highbit
      shr eax, 31-highbit+lowbit
      EXITM <eax>
    ENDM

; -------------------------------------------------------------

.data
    inbuf       db 130 dup(0)
    outbuf      db 30 dup(0)
    crlf        db 13, 10, 0
    NL          dw offset crlf
    backspace   db 8, 0
    BS          dw OFFSET backspace
    hexChars    db "0123456789ABCDEF"
.code

; -------------------------------------------------------------
; This proc calls the Read File or Device function to read
; up to <blen> bytes from STDIN and store them in the
; specified buffer.
; -------------------------------------------------------------

stdin proc c uses bx cx dx npdest:WORD, blen:WORD
    mov bx, HSTDIN
    mov cx, blen
    mov dx, npdest
    mov ah, 3fh
    int 21h
    mov bx, npdest
  @@:                         ; Replace CR with null
    cmp BYTE PTR [bx], 0
    je  @F
    cmp BYTE PTR [bx], 13
    je  @F
    inc bx
    jmp @B
  @@:
    mov BYTE PTR [bx], 0
    ret
stdin endp

; -------------------------------------------------------------
; This proc returns the length of the specified
; null-terminated string in AX.
; -------------------------------------------------------------

szlen proc c uses bx npsz:WORD
    mov bx, npsz
  runLoop:
    cmp BYTE PTR [bx], 0      ; Compare bytes until find null
    jz  @F
    inc bx
    jmp runLoop
  @@:
    mov ax, bx                ; Calc length and return in AX
    sub ax, npsz
    ret
szlen endp

; -------------------------------------------------------------
; This proc calls the Write File or Device function to
; output the specified null-terminated string to STDOUT.
; -------------------------------------------------------------

stdout proc c uses ax bx cx dx npsz:WORD
    invoke szlen, npsz
    mov cx, ax
    mov bx, HSTDOUT
    mov dx, npsz
    mov ah, 40h
    int 21h
    ret
stdout endp

; -------------------------------------------------------------
; This proc calls the Display Character function to output
; the specified character to STDOUT.
; -------------------------------------------------------------

putchar proc c uses ax dx char:BYTE
    mov ah, 2
    mov dl, char
    int 21h
    ret
putchar endp

; -------------------------------------------------------------
; This proc calls the BIOS Read Keyboard Status and Read
; Keyboard Input functions to flush the keyboard buffer,
; wait for a key press, and then flush the buffer again,
; leaving it empty. The BIOS scan code for the key pressed
; is returned in AH and the character code in AL.
; -------------------------------------------------------------

waitkey proc c
  @@:
    mov ah, 1
    int 16h                   ; Check for key press
    jz  @F                    ; Jump if buffer empty
    mov ah, 0
    int 16h                   ; Get key
    jmp @B                    ; Repeat
  @@:
    mov ah, 1
    int 16h                   ; Check for key press
    jz  @B                    ; Repeat
    mov ah, 0
    int 16h                   ; Get key
    push ax                   ; Preserve it
  @@:
    mov ah, 1
    int 16h                   ; Check for key press
    jz  @F                    ; Jump if buffer empty
    mov ah, 0
    int 16h                   ; Get key
    jmp @B                    ; Repeat
  @@:
    pop ax                    ; Recover key
    ret
waitkey endp

; -------------------------------------------------------------
; This proc calls the BIOS Set Cursor Position function to
; set the cursor position for display page 0 to <row>,<col>
; (upper left corner is 0,0).
; -------------------------------------------------------------

locate proc c uses ax bx dx row:BYTE, col:BYTE
    mov ah, 2
    mov bh, 0                 ; Display page
    mov dh, row
    mov dl, col
    int 10h
    ret
locate endp

; -------------------------------------------------------------
; These procs return the row and column cursor positions
; for display page 0 in AL.
; -------------------------------------------------------------

cursorrow proc c uses es
    push 40h                  ; Set ES to the BIOS data area
    pop es
    mov al, es:[51h]          ; Read the appropriate byte
    ret
cursorrow endp

cursorcol proc c uses es
    push 40h
    pop es
    mov al, es:[50h]
    ret
cursorcol endp

; -------------------------------------------------------------
; This proc clears the screen by calling the BIOS Scroll
; Current Page Up function with a scroll (row) count of zero,
; and sets the cursor position to the upper left corner of
; the screen. An 80x25 text mode screen is assumed.
; -------------------------------------------------------------

_cls proc c uses ax bx cx dx
    mov ah, 6                 ; Function number
    mov al, 0                 ; Row count
    mov bh, 7                 ; Attribute for blanked lines
                              ; (white on black)
    mov ch, 0                 ; Top row of window
    mov cl, 0                 ; Left-most column of window
    mov dh, 24                ; Bottom row of window
    mov dl, 79                ; Right-most column of window
    int 10h
    invoke locate, 0, 0
    ret
_cls endp

; -------------------------------------------------------------
; This proc converts the specified null-terminated decimal
; string to a signed 32-bit integer value and returns the
; result in EAX. The procedure stops reading the string at the
; first non-numeric character, which may be the terminating
; null. This procedure cannot correctly handle values outside
; the range -2147483648 to 2147483647.
; -------------------------------------------------------------

atodw proc c uses dx si npsz:WORD
    LOCAL fNeg:WORD

    mov si, npsz
    xor edx, edx              ; Use EDX to store value
  @@:
    lodsb
    cmp al, 32
    je  @B                    ; Scan off leading spaces
    cmp al, 9
    je  @B                    ; Scan off leading tabs
    cmp al, '+'
    je  @B                    ; Scan off leading '+'
    cmp al, '-'               ; Check for leading '-'
    jne @F
    mov fNeg, 1
    jmp digitLoop
  @@:
    mov fNeg, 0               ; Char is digit, clear flag and
    dec si                    ;  decrement SI to repeat char
  digitLoop:
    lodsb
    cmp al, '0'
    jb  endOfString           ; Stop at non-numeric char
    cmp al, '9'
    ja  endOfString           ; Stop at non-numeric char
    cmp edx, 0
    jne @F
    cmp al, '0'
    je  digitLoop             ; Ignore any leading zeros
  @@:
    sub al, '0'               ; Convert to decimal value
    cmp edx, 0
    je  @F                    ; If not first digit
    imul edx, 10              ;  multiply total times base
  @@:
    movzx eax, al             ; Zero extend AL into EAX
    add edx, eax              ; Add in value for digit
    jmp digitLoop
  endOfString:
    mov eax, edx
    cmp fNeg, 0
    je  @F
    neg eax
  @@:
    ret
atodw endp

; -------------------------------------------------------------
; This proc converts a byte to a signed null-terminated
; decimal string and stores the result (4 characters max)
; in the specified buffer.
;
; This proc cannot correctly handle values that are outside
; the range -128 to 127.
; -------------------------------------------------------------

b2dec proc c uses ax bx cx dx di src:BYTE, npdest:WORD
    mov al, src
    mov di, npdest
    test al, al               ; Test for negative value
    jns @F                    ; Make positive if negative
    neg al
    mov BYTE PTR [di], '-'    ; Store sign in buffer
    inc di                    ; Adjust DI for next character
  @@:
    mov bl, 10                ; Init divisor
    xor cx, cx                ; Zero digit counter
  divLoop:
    xor ah, ah                ; Zero MSB of dividend
    div bl                    ; Divide value by 10
    movzx dx, ah
    push dx                   ; Push remainder
    inc cx                    ; Increment digit count
    or  al, al                ; Repeat if quotient > 0
    jnz divLoop
  storeLoop:
    pop ax                    ; Pop digit off stack
    add al, '0'               ; Convert to ascii
    mov [di], al              ; Store in buffer
    inc di                    ; Adjust DI for next character
    loop storeLoop            ; Repeat CX times
    mov BYTE PTR [di], 0      ; Append trailing null
    ret
b2dec endp

; -------------------------------------------------------------
; This proc converts a byte to a null-terminated 2-character
; hex string and stores the result in the specified buffer.
; -------------------------------------------------------------

;THIS IS PROBLEM FOR ML 7.00
  ;.data
  ;  hexChars db "0123456789ABCDEF"
  ;.code

b2hex proc c uses ax bx dx di src:BYTE, npdest:WORD
    mov dl, src
    mov di, npdest
    add di, 1                 ; Set DI to end of buffer
    mov BYTE PTR [di+1], 0    ; Append trailing null
  @@:
    movzx bx, dl              ; Copy current value into BX
    and bx, 0Fh               ; Mask off all but lower nibble
    mov al, hexChars[bx]      ; Load corresponding char into AL
    mov [di], al              ; Store in buffer
    shr dl, 4                 ; Shift next nibble into position
    dec di                    ; Adjust DI for next character
    cmp di, npdest            ; Repeat for all 2 characters
    jnb @B
    ret
b2hex endp

; -------------------------------------------------------------
; This proc converts a word to a signed null-terminated
; decimal string and stores the result (6 characters max)
; in the specified buffer.
;
; This proc cannot correctly handle values that are outside
; the range -32768 to 32767.
; -------------------------------------------------------------

w2dec proc c uses ax bx cx dx di src:WORD, npdest:WORD
    mov ax, src
    mov di, npdest
    test ax, ax               ; Test for negative value
    jns @F                    ; Make positive if negative
    neg ax
    mov BYTE PTR [di], '-'    ; Store sign in buffer
    inc di                    ; Adjust DI for next character
  @@:
    mov bx, 10                ; Init divisor
    xor cx, cx                ; Zero digit counter
  divLoop:
    xor dx, dx                ; Zero MSW of dividend
    div bx                    ; Divide value by 10
    push dx                   ; Push remainder
    inc cx                    ; Increment digit count
    or  ax, ax                ; Repeat if quotient > 0
    jnz divLoop
  storeLoop:
    pop ax                    ; Pop digit off stack
    add al, '0'               ; Convert to ascii
    mov [di], al              ; Store in buffer
    inc di                    ; Adjust DI for next character
    loop storeLoop            ; Repeat CX times
    mov BYTE PTR [di], 0      ; Append trailing null
    ret
w2dec endp

; -------------------------------------------------------------
; This proc converts a word to a null-terminated 4-character
; hex string and stores the result in the specified buffer.
; -------------------------------------------------------------

w2hex proc c uses ax bx dx di src:WORD, npdest:WORD
    mov dx, src
    mov di, npdest
    add di, 3                 ; Set DI to end of buffer
    mov BYTE PTR [di+1], 0    ; Append trailing null
  @@:
    mov bx, dx                ; Copy current value into BX
    and bx, 0Fh               ; Mask off all but lower nibble
    mov al, hexChars[bx]      ; Load corresponding char into AL
    mov [di], al              ; Store in buffer
    shr dx, 4                 ; Shift next nibble into position
    dec di                    ; Adjust DI for next character
    cmp di, npdest            ; Repeat for all 4 characters
    jnb @B
    ret
w2hex endp

; -------------------------------------------------------------
; This proc converts a dword to an signed null-terminated
; decimal string and stores the result (11 characters max)
; in the specified buffer.
;
; This proc cannot correctly handle values that are outside
; the range -2147483648 to 2147483647.
; -------------------------------------------------------------

dw2dec proc c uses ax bx cx dx di src:DWORD, npdest:WORD
    mov eax, src
    mov di, npdest
    test eax, eax             ; Test for negative value
    jns @F
    neg eax                   ; Make positive if negative
    mov BYTE PTR[di], '-'     ; Store sign in buffer
    inc di
  @@:
    mov ebx, 10               ; Init divisor
    xor cx, cx                ; Zero digit counter
  divLoop:
    xor edx, edx              ; Zero MSDW of dividend
    div ebx                   ; Divide value by 10
    push dx                   ; Push remainder
    inc cx                    ; Increment digit count
    or  eax, eax              ; Repeat if quotient > 0
    jnz divLoop
  storeLoop:
    pop ax                    ; Pop digit off stack
    add al, '0'               ; Convert to ascii
    mov [di], al              ; Store in buffer
    inc di                    ; Adjust DI for next character
    loop storeLoop            ; Repeat CX times
    mov BYTE PTR [di], 0      ; Append trailing null
    ret
dw2dec endp

; -------------------------------------------------------------
; This proc converts a dword to a null-terminated
; 8-character hex string and stores the result in the
; specified buffer.
; -------------------------------------------------------------

dw2hex proc c uses ax bx dx di src:DWORD, npdest:WORD
    mov edx, src
    mov di, npdest
    add di, 7                 ; Set DI to end of buffer
    mov BYTE PTR [di+1], 0    ; Append trailing null
  @@:
    mov ebx, edx              ; Copy current value into EBX
    and ebx, 0Fh              ; Mask off all but lower nibble
    mov al, hexChars[ebx]     ; Load corresponding char into AL
    mov [di], al              ; Store in buffer
    shr edx, 4                ; Shift next nibble into position
    dec di                    ; Adjust DI for next character
    cmp di, npdest
    jnb @B                    ; Repeat for all 8 characters
    ret
dw2hex endp

; -------------------------------------------------------------
; This proc converts a byte to to a null-terminated
; 8-character binary string and stores the result in
; the specified buffer.
; -------------------------------------------------------------

b2bin proc c uses ax bx cx dx src:BYTE, npdest:WORD
    mov al, src
    mov bx, npdest
    mov cx, 8
    dec bx                    ; DEC BX so first INC in loop
  shiftLoop:                  ;  will not skip first bit
    inc bx
    xor dx, dx                ; Zero DX
    shl al, 1                 ; Shift next bit into carry flag
    adc dl, '0'               ; Copy carry flag to DL and
                              ;  convert to '0' or '1'
    mov [bx], dl              ; Store in buffer
    dec cx
    jnz shiftLoop             ; Repeat for all bits
    mov BYTE PTR [bx+1], 0    ; Append the terminating null
    ret
b2bin endp

; -------------------------------------------------------------
; This proc converts any upper-case characters in the
; specified null-terminated string to lower case.
; -------------------------------------------------------------

szlower proc c uses ax bx npsz:WORD
    mov bx, npsz
  runLoop:
    mov al, [bx]
    test al, al               ; Test for terminating null
    jnz @F
    ret                       ; Have reached end of string
  @@:
    cmp al, 'A'               ; Convert upper-case characters
    jb  @F                    ;  only
    cmp al, 'Z'               ;
    ja  @F                    ;
    add al, 'a'-'A'           ;
    mov [bx], al              ;
  @@:
    inc bx
    jmp   runLoop
szlower endp

; -------------------------------------------------------------
; This proc converts any lower-case characters in the
; specified null-terminated string to upper case.
; -------------------------------------------------------------

szupper proc c uses ax bx npsz:WORD
    mov bx, npsz
  runLoop:
    mov al, [bx]
    test al,al                ; Test for terminating null
    jnz @F
    ret                       ; Have reached end of string
  @@:
    cmp al, 'a'               ; Convert lower-case characters
    jb  @F                    ;  only
    cmp al, 'z'               ;
    ja  @F                    ;
    sub al, 'a'-'A'           ;
    mov [bx], al              ;
  @@:
    inc bx
    jmp   runLoop
szupper endp

; -------------------------------------------------------------
; This proc displays the value of the Overflow, Direction,
; Interrupt Enable, Sign, Zero, Auxiliary Carry, Parity, and
; Carry flags, in a DEBUG-style format.
;
;   Flag          position  flag set flag clear
;   ----          --------  -------- ----------
; Overflow        (bit 11)     OV        NV
; Direction       (bit 10)     DN        UP
; Interrupt       (bit 9)      EI        DI
; Sign            (bit 7)      NG        PL
; Zero            (bit 6)      ZR        NZ
; Auxiliary Carry (bit 4)      AC        NA
; Parity          (bit 2)      PE        PO
; Carry           (bit 0)      CY        NC
; -------------------------------------------------------------

dumpflags proc
    pushf
    pushad
    pushf
    pop   ax
    .IF ax & 1 SHL 11         ; Overflow (bit 11)
      print "OV "
    .ELSE
      print "NV "
    .ENDIF
    .IF ax & 1 SHL 10         ; Direction (bit 10)
      print "DN "
    .ELSE
      print "UP "
    .ENDIF
    .IF ax & 1 SHL 9          ; Interrupt (bit 9)
      print "EI "
    .ELSE
      print "DI "
    .ENDIF
    .IF ax & 1 SHL 7          ; Sign (bit 7)
      print "NG "
    .ELSE
      print "PL "
    .ENDIF
    .IF ax & 1 SHL 6          ; Zero (bit 6)
      print "ZR "
    .ELSE
      print "NZ "
    .ENDIF
    .IF ax & 1 SHL 4          ; Auxiliary Carry (bit 4)
      print "AC "
    .ELSE
      print "NA "
    .ENDIF
    .IF ax & 1 SHL 2          ; Parity (bit 2)
      print "PE "
    .ELSE
      print "PO "
    .ENDIF
    .IF ax & 1 SHL 0          ; Carry (bit 0)
      print "CY ",NL
    .ELSE
      print "NC ",NL
    .ENDIF
    popad
    popf
    ret
dumpflags endp

; -------------------------------------------------------------
; This proc displays a hex-ascii dump of the specified memory.
; -------------------------------------------------------------

dumpmem proc uses es mem_seg:WORD, mem_off:WORD, paragraphs:WORD
    pushf
    pushad
    mov es, mem_seg
    mov si, mem_off
    mov cx, paragraphs
    test cx, cx
    jz  fini
  addressloop:
    print hexword$(es),":"
    print hexword$(si),"  "
    xor bx, bx
  hexloop:
    print hexbyte$(es:[si+bx])," "
    cmp bx, 7
    jne @F
    print "- "
  @@:
    inc bx
    cmp bx, 16
    jb  hexloop
    xor bx, bx
    putch ' '
  asciiloop:
    mov al, es:[si+bx]
    cmp al, 32
    jna badchar               ; bad char if < 33
    cmp al, 127
    jnb badchar               ; bad char if > 126
    putch al
    jmp @F
  badchar:
    putch '.'                 ; substitute a period
  @@:
    inc bx
    cmp bx, 16
    jb  asciiloop
    print NL
    add si, 16
    dec cx
    jnz addressloop
  fini:
    popad
    popf
    ret
dumpmem endp

; -------------------------------------------------------------
; This proc returns the BIOS maintained count of timer ticks
; since midnight in EAX. The timer ticks are generated by
; PIT channel 0, normally programed by the BIOS to generate
; 1193182/65536 = 18.207 ticks per second.
; -------------------------------------------------------------

get_tick_count proc uses es bx
    mov ax, 40h               ; BIOS data area segment address
    mov es, ax
    mov bx, 6ch
    mov eax, es:[bx]
    ret
get_tick_count endp

; -------------------------------------------------------------
; This proc delays for the specified number of memory refresh
; requests. The requests are generated by PIT channel 1,
; normally programmed by the BIOS to generate an output once
; every 18/1193182 = 15.09 microseconds. Each request toggles
; bit 4 of I/O port 61h.
; -------------------------------------------------------------

delay proc uses ax ecx dwcount:DWORD
    mov ecx, dwcount
  waituntilset:
    in  al, 61h
    test al, 10h
    jz  waituntilset
  waituntilclear:
    in  al, 61h
    test al, 10h
    jnz waituntilclear
    sub ecx, 1
    jnz waituntilset
    ret
delay endp

; -------------------------------------------------------------
; This proc returns the value from the specified CMOS address
; in AL.
;
; Bit 7 of the CMOS address port is used to control the
; state of the NMI (Nonmaskable Interrupt). The remaining
; 7 bits are used to specify one of 128 possible addresses.
;
; The state of the NMI disable cannot be determined, so this
; proc assumes that the NMI was enabled, and re-enables it
; before exit.
;
; For the time and date registers this procedure implements
; the logic necessary to reliably read these registers.
;
; In practice, for anything resembling a recent system, the
; I/O delay is probably not necessary because the chipset
; inserts its own delay.
;
; This proc is based mostly on the read_CMOS example, and
; related information, from Frank van Gilluwe's book The
; Undocumented PC, First Edition, Addison-Wesley, 1994,
; ISBN 0-201-62277-7.
; -------------------------------------------------------------

IODELAY MACRO
    REPEAT 10
      jmp   $+2
    ENDM
ENDM

read_cmos proc c address:BYTE
    cli                       ; Disable hardware interrupts.
    .IF address < 0Ah
        mov al, 0Ah           ; Address of status register A.
        or  al, 80h           ; Set bit 7 to disable NMI.
        out 70h, al           ; Write to address port.
      @@:
        IODELAY
        in  al, 71h           ; Wait until update-in-progress
        test al, 80h          ;  bit (bit 7) is set.
        jz  @B
        mov al, 0Ah           ; This second write is necessary
        out 70h, al           ;  for some systems.
      @@:
        IODELAY
        in  al, 71h           ; Wait until update-in-progress
        test al, 80h          ;  bit is clear.
        jnz @B
    .ENDIF
    mov al, address
    or  al, 80h               ; Set bit 7 to disable NMI.
    out 70h, al               ; Write to address port.
    IODELAY
    in  al, 71h               ; Read value from data port.
    IODELAY
    mov ah, al                ; Preserve value.
    mov al, 0Dh               ; Set the address port to status
    out 70h, al               ;  register D and enable NMI.
    in  al, 71h               ; Do dummy read.
    shr ax, 8                 ; Set return value.
    sti                       ; Enable hardware interrupts.
    ret
read_cmos endp

; -------------------------------------------------------------
; This proc outputs the specified far string, enclosing it
; in single quotes.
; -------------------------------------------------------------

print_far_string proc c uses es bx str_seg:WORD, str_ptr:WORD
    LOCAL char[2]:BYTE

    print "'"
    mov char[1], 0            ; Initialize null terminator
    mov es, str_seg
    mov bx, str_ptr
    .WHILE BYTE PTR es:[bx] != 0
      mov al, es:[bx]
      mov char, al
      print ADDR char
      inc bx
    .ENDW
    print "'"
    ret
print_far_string endp

; -------------------------------------------------------------
